import sys
import os
import gzip
import re
import logging
import click_log
import collections
import re
from collections import OrderedDict
from math import ceil, floor

from construct import *
from inspect import getframeinfo, currentframe, getdoc

import pysam

from ..meta import VERSION
from ..utils.model import reverse_complement

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("bam_utils")
click_log.basic_config(logger)

PB_READ_NAME_RE = re.compile("m[0-9]+e?_[0-9]{6}_[0-9]{6}/[0-9]+/.*")

# Constants for bam file reading / writing:
SEGMENTS_TAG = "SG"
SEGMENTS_QUAL_TAG = "XQ"
SEGMENTS_RC_TAG = "RC"
SEGMENT_TAG_DELIMITER = ","
READ_MODEL_NAME_TAG = "YN"
READ_MODEL_SCORE_TAG = "YS"
READ_IS_VALID_FOR_MODEL_TAG = "YV"
READ_FIRST_KEY_SEG_TAG = "YK"
READ_NUM_KEY_SEGMENTS_TAG = "YG"
READ_APPROX_QUAL_TAG = "YQ"

READ_UMI_TAG = ""


# Named tuple to store alignment information:
class SegmentInfo(collections.namedtuple("SegmentInfo", ["name", "start", "end"])):

    _tag_regex = re.compile(r"(.*?):(\d+)-(\d+)")

    def __len__(self):
        return self.end - self.start

    def __str__(self):
        return f"SegmentInfo({self.to_tag()})"

    def to_tag(self):
        return f"{self.name}:{self.start}-{self.end}"

    @classmethod
    def from_tag(cls, tag_string):
        match = cls._tag_regex.match(tag_string)
        return SegmentInfo(match[1], int(match[2]), int(match[3]))


def load_read_count(pbi_file):
    """Compute file offsets for specified read names"""

    # Decode PacBio .pbi file.  This is not a full decode of the index, only the parts we need
    # until we get to the read count.
    # More on index format at https://pacbiofileformats.readthedocs.io/en/9.0/PacBioBamIndex.html .

    fmt = Struct(
        # Header
        "magic" / Const(b"PBI\x01"),
        "version_patch" / Int8ul,
        "version_minor" / Int8ul,
        "version_major" / Int8ul,
        "version_empty" / Int8ul,
        "pbi_flags" / Int16ul,
        "n_reads" / Int32ul,
        )

    with gzip.open(pbi_file, "rb") as f:
        idx_contents = fmt.parse_stream(f)

        return idx_contents.n_reads


def compute_shard_offsets(pbi_file, num_shards):
    """
    Compute all possible shard offsets (keeping adjacent reads from the ZMW together)
    """

    # Decode PacBio .pbi file.  This is not a full decode of the index, only the parts we need for sharding.
    # More on index format at https://pacbiofileformats.readthedocs.io/en/9.0/PacBioBamIndex.html .

    fmt = Struct(
        # Header
        "magic" / Const(b"PBI\x01"),
        "version_patch" / Int8ul,
        "version_minor" / Int8ul,
        "version_major" / Int8ul,
        "version_empty" / Int8ul,
        "pbi_flags" / Int16ul,
        "n_reads" / Int32ul,
        "reserved" / Padding(18),

        # Basic information section (columnar format)
        "rgId" / Padding(this.n_reads * 4),
        "qStart" / Padding(this.n_reads * 4),
        "qEnd" / Padding(this.n_reads * 4),
        "holeNumber" / Array(this.n_reads, Int32sl),
        "readQual" / Padding(this.n_reads * 4),
        "ctxtFlag" / Padding(this.n_reads * 1),
        "fileOffset" / Array(this.n_reads, Int64sl),
        )

    # Make a list of bgzf virtual file offsets for sharding and store ZMW counts.
    file_offsets_hash = OrderedDict()
    last_offset = 0
    zmw_count_hash = dict()
    with gzip.open(pbi_file, "rb") as f:
        idx_contents = fmt.parse_stream(f)

        for j in range(0, idx_contents.n_reads):
            # Save only the virtual file offset for the first ZMW hole number, so
            # that shard boundaries always keep reads from the same ZMW together.
            if idx_contents.holeNumber[j] not in file_offsets_hash:
                file_offsets_hash[idx_contents.holeNumber[j]] = idx_contents.fileOffset[j]

            last_offset = idx_contents.fileOffset[j]

            try:
                zmw_count_hash[idx_contents.holeNumber[j]] += 1
            except KeyError:
                zmw_count_hash[idx_contents.holeNumber[j]] = 1

    file_offsets = list(file_offsets_hash.values())
    shard_offsets = []
    read_counts = []
    for j in range(0, len(file_offsets), ceil(len(file_offsets) / num_shards)):
        shard_offsets.append(file_offsets[j])
        read_counts.append(ceil(len(file_offsets) / num_shards))

    # For the last read in the file, pad the offset so the final comparison in write_shard() retains the final read.
    shard_offsets.append(os.path.getsize(pbi_file) + 1)

    return shard_offsets, zmw_count_hash, idx_contents.n_reads, read_counts


def create_bam_header_with_program_group(command_name, base_bam_header, description=None, models=None):
    """Create a pysam.AlignmentHeader object with program group (PG) information populated by the given arguments.

    This function is intended to be called from the 'main' function of a longbow subcommand because it uses reflection
    to pull in the first line of the docstring from the main function as the description (DS field)."""

    bam_header_dict = base_bam_header.to_dict()

    if not description:
        prev_frame = currentframe().f_back
        description = getdoc(prev_frame.f_globals['main']).split("\n")[0]

    # If we have a model here, we should add the description of the model to our program group:
    if models:
        description = description + "  MODEL(s): " + ", ".join([m.to_json(indent=None) for m in models])

    # Add our program group to it:
    pg_dict = {
        "ID": f"longbow-{command_name}-{VERSION}",
        "PN": "longbow",
        "VN": f"{VERSION}",
        # Use reflection to get the first line of the doc string the caller - the main function for our header:
        "DS": description,
        "CL": " ".join(sys.argv),
    }
    if "PG" in bam_header_dict:
        bam_header_dict["PG"].append(pg_dict)
    else:
        bam_header_dict["PG"] = [pg_dict]
    out_header = pysam.AlignmentHeader.from_dict(bam_header_dict)

    return out_header


def check_for_preexisting_files(file_list, exist_ok=False):
    """Checks if the files in the given file_list exist.
    If any file exists and exist_ok is False, this will exit the program.
    If any file exists and exist_ok is True, the program will continue.
    """

    # Allow users to be a little lazy with what input types they give:
    if not isinstance(file_list, list) and not isinstance(file_list, set):
        file_list = [file_list]

    do_files_exist = False
    for f in file_list:
        if os.path.exists(f):
            if exist_ok:
                logger.warning(f"Output file exists: {f}.  Overwriting.")
            else:
                logger.error(f"Output file already exists: {f}!")
                do_files_exist = True
    if do_files_exist:
        sys.exit(1)


def get_segment_score(read_sequence, segment, model, ssw_aligner=None):
    """Get the alignment score of the given segment against the read sequence."""

    # We don't score random segments:
    if segment.name == "random":
        return 0, 0

    # Create a default aligner if we weren't given one:
    if not ssw_aligner:
        ssw_aligner = ssw.Aligner()

    # Get our alignment and our score:
    if segment.end - segment.start > 1:
        alignment = ssw_aligner.align(read_sequence[segment.start:segment.end], model.adapter_dict[segment.name])
        optimal_score = alignment.score
    else:
        optimal_score = 0

    # The max score is the match score * the length of the reference segment
    max_score = len(model.adapter_dict[segment.name]) * ssw_aligner.matrix.get_match()

    return optimal_score, max_score


def collapse_annotations(path):
    """Collapses given path into a list of SegmentInfo objects."""
    last = ""
    start = 0
    segments = []
    i = 0
    for i, seg in enumerate(path):
        if seg != last:
            if i != 0:
                segments.append(SegmentInfo(last, start, i - 1))
            last = seg
            start = i
    # Don't forget the last one:
    segments.append(SegmentInfo(last, start, i))

    return segments


def write_annotated_read(read, segments, is_rc, logp, model, ssw_aligner, out_bam_file):
    """Write the given pysam.AlignedSegment read object to the given file with the given metadata."""

    # Obligatory log message:
    logger.debug(
        "Path for read %s (%2.2f)%s: %s",
        read.query_name,
        logp,
        " (RC)" if is_rc else "",
        segments,
    )

    # Set our tag and write out the read to the annotated file:
    read.set_tag(SEGMENTS_TAG, SEGMENT_TAG_DELIMITER.join([s.to_tag() for s in segments]))

    # Set the model info tags:
    read.set_tag(READ_MODEL_SCORE_TAG, logp)
    read.set_tag(READ_MODEL_NAME_TAG, model.name)

    # If we're reverse complemented, we make it easy and just reverse complement the read and add a tag saying
    # that the read was RC:
    read.set_tag(SEGMENTS_RC_TAG, is_rc)
    if is_rc:
        quals = read.query_qualities[::-1]
        seq = reverse_complement(read.query_sequence)
        read.query_sequence = seq
        read.query_qualities = quals

    # Get our segment scores and set them:
    total_score = 0
    total_max_score = 0
    score_strings = []
    for s in segments:
        score, max_score = get_segment_score(read.query_sequence, s, model, ssw_aligner)
        score_strings.append(f"{score}/{max_score}")
        total_score += score
        total_max_score += max_score

    read.set_tag(SEGMENTS_QUAL_TAG, SEGMENT_TAG_DELIMITER.join(score_strings))
    if total_max_score != 0:
        read.set_tag(READ_APPROX_QUAL_TAG, f"{total_score / total_max_score:.4f}")
    else:
        read.set_tag(READ_APPROX_QUAL_TAG, f"0.0")

    out_bam_file.write(read)



import logging
import time
import os
import sys

import click
import click_log

import tqdm
import pysam
from construct import *

from ..utils import bam_utils
from ..annotate.command import get_segments

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("extract")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(exists=True),
    help="BAM .pbi index file",
)
@click.option(
    "-o",
    "--out-file",
    default=".",
    required=True,
    type=str,
    help="Output file name.",
)
@click.option(
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.option(
    "-b",
    "--base-padding",
    default=2,
    required=False,
    show_default=True,
    type=int,
    help="Number of bases to include on either side of the extracted region(s).",
)
@click.option(
    "--leading-adapter",
    default="10x_Adapter",
    required=False,
    show_default=True,
    type=str,
    help="Adapter preceding the region to extract.",
)
@click.option(
    "--trailing-adapter",
    default="Poly_A",
    required=False,
    show_default=True,
    type=str,
    help="Adapter following the region to extract.",
)
@click.option(
    "--start-offset",
    default=16+10,  # CBC + UMI for MAS15 is the default
    required=False,
    show_default=True,
    type=int,
    help="Number of bases to ignore from the extracted region start.  "
         "These bases will not be included in the extracted sequences.",
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, out_file, force, base_padding, leading_adapter, trailing_adapter, start_offset, input_bam):
    """Extract coding segments from the reads in the given bam.
    The main coding segments are assumed to be labeled as `random` segments.
    Uses known segments flanking the region to be extracted as markers to indicate
    the start and end of what to extract."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    read_count = None
    if os.path.exists(pbi):
        read_count = bam_utils.load_read_count(pbi)
        logger.info("About to Extract segments from %d reads", read_count)

    # Check to see if the output file exists:
    bam_utils.check_for_preexisting_files(out_file, exist_ok=force)

    # TODO: We don't need to check our model right now because our models are SO similar.
    #       This will need to be fixed when we start using more exotic models.

    logger.info(f"Writing extracted read segments to: {out_file}")
    logger.info(f"Extracting `random` segments between {leading_adapter} and {trailing_adapter}.")
    logger.info(f"Ignoring the first {start_offset} bases from extracted read segments.")
    logger.info(f"Including {base_padding} flanking bases.")

    # Open our input bam file:
    pysam.set_verbosity(0)
    with pysam.AlignmentFile(input_bam, "rb", check_sq=False, require_index=False) as bam_file, \
            tqdm.tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            disable=not sys.stdin.isatty(),
            total=read_count
            ) as pbar:

        # Get our header from the input bam file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header)

        # Setup output files:
        with pysam.AlignmentFile(out_file, "wb", header=out_header) as extracted_bam_file:

            num_reads = 0
            num_reads_with_extracted_segments = 0
            num_segments_extracted = 0
            num_segments_skipped = 0

            for read in bam_file:
                # Get our read segments:
                try:
                    _, segments = get_segments(read)
                except KeyError:
                    logger.error(f"Input bam file does not contain longbow segmented reads!  "
                                 f"No {bam_utils.SEGMENTS_TAG} tag detected on read {read.query_name} !")
                    sys.exit(1)

                # Get our marker segments:
                start_marker_list = [(i, s) for i, s in enumerate(segments) if s.name == leading_adapter]
                end_marker_list = [(i, s) for i, s in enumerate(segments) if s.name == trailing_adapter]

                if len(start_marker_list) != len(end_marker_list):
                    logger.warning(f"Found %d start markers and %d end markers.  Only looking at first %d pairs.  "
                                   f"(starts: %s, ends: %s)",
                                   len(start_marker_list), len(end_marker_list),
                                   min(len(start_marker_list), len(end_marker_list)),
                                   " ".join([f"{i}:{s.name}" for i, s in start_marker_list]),
                                   " ".join([f"{i}:{s.name}" for i, s in end_marker_list]))

                extracted_segment = False

                # Go through each marker pair and do the extraction:
                for s_info, e_info in zip(start_marker_list, end_marker_list):

                    si, start_marker = s_info
                    ei, end_marker = e_info

                    # Does the start marker come before the end marker and do we have exactly one random segment in
                    # between them?
                    if (start_marker.end < end_marker.start) and (ei - si == 2) and (segments[si + 1].name == "random"):
                        # We have a valid segment to extract:
                        logger.debug("Found a segment to extract: %s: %s", read.query_name, segments[si + 1])

                        # Create an AlignedSegment to output:
                        aligned_segment = _create_extracted_aligned_segment(
                            read, segments[si + 1], start_offset, base_padding
                        )

                        if aligned_segment:
                            extracted_bam_file.write(aligned_segment)
                            num_segments_extracted += 1
                            extracted_segment = True
                        else:
                            num_segments_skipped += 1
                    else:
                        if start_marker.end >= end_marker.start:
                            logger.warning("Read %s: start marker segment (i=%d) occurs at or after end segment (i=%d):"
                                           " %d >= %d.  Skipping segment.",
                                           read.query_name, si, ei, start_marker.end, end_marker.start)
                        elif ei - si != 2:
                            logger.warning("Read %s: start segment (i=%d) and end segment (i=%d) have more than one "
                                           "segment between them.  Skipping segment.", read.query_name, si, ei)
                        elif segments[si + 1].name != "random":
                            logger.warning("Read %s: segment between start segment (i=%d) and end segment (i=%d) "
                                           "is not a random segment.  Skipping segment.", read.query_name, si, ei)
                        num_segments_skipped += 1

                pbar.update(1)
                num_reads += 1
                if extracted_segment:
                    num_reads_with_extracted_segments += 1

    # Calc some stats:
    pct_reads_with_extracted_segments = 100 * num_reads_with_extracted_segments / num_reads if num_reads > 0 else 0
    segs_per_read = num_segments_extracted / num_reads if num_reads > 0 else 0

    # Yell at the user:
    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)
    logger.info(f"Total # Reads Processed: %d", num_reads)
    logger.info(f"# Reads Containing Extracted Segments: %d (%2.2f%%)",
                num_reads_with_extracted_segments, pct_reads_with_extracted_segments)
    logger.info(f"Total # Segments Extracted: %d", num_segments_extracted)
    logger.info(f"Total # Segments Skipped: %d", num_segments_skipped)
    logger.info(f"# Segments extracted per read: %2.2f", segs_per_read)


def _create_extracted_aligned_segment(read, seg_to_extract, start_offset, base_padding):
    """Create a pysam.AlignedSegment object to store the information from the extracted bases."""

    start_coord = seg_to_extract.start + start_offset - base_padding
    end_coord = seg_to_extract.end + base_padding

    # Bounds check our coords:
    if start_coord < 0:
        logger.debug("Calculated start for %s would start before read begins.  Setting to 0.", read.query_name)
        start_coord = 0
    elif start_coord >= len(read.query_sequence):
        logger.warning("Start coord for %s would start after read.  Cannot process.", read.query_name)
        return None

    if end_coord < 0:
        logger.warning("End coord for %s would start before read.  Cannot process.", read.query_name)
        return None
    elif end_coord >= len(read.query_sequence):
        logger.debug("Calculated end for %s would start after read ends.  Setting to 0.", read.query_name)
        end_coord = len(read.query_sequence)-1

    if end_coord <= start_coord:
        logger.warning("Start coord for %s would start at or after end coord.  Cannot process.", read.query_name)
        return None

    # Create our segment:
    a = pysam.AlignedSegment()
    a.query_name = (
        f"{read.query_name}/{start_coord}_{end_coord}"
    )
    # Add one to end_coord because coordinates are inclusive:
    a.query_sequence = f"{read.query_sequence[start_coord:end_coord+1]}"
    a.query_qualities = read.query_alignment_qualities[start_coord: end_coord + 1]
    a.tags = read.get_tags()
    a.flag = 4  # unmapped flag
    a.mapping_quality = 255

    return a


import sys
import re
import json
import logging

import click_log

import numpy as np
import pandas as pd

from pomegranate import *
from pomegranate.callbacks import History, ModelCheckpoint


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger(__name__)
click_log.basic_config(logger)


class LibraryModel:
    """Model describing a given library preparation.
    The model can annotate the known sections of a read from the library it describes."""

    # Define constants for all our default probabilities here:
    RANDOM_BASE_PROB = 0.5
    PER_BASE_MATCH_PROB = 0.94
    PER_BASE_MISMATCH_PROB = 0.02

    MATCH_MATCH_PROB = 0.90
    MATCH_INDEL_PROB = 0.05
    MATCH_TRAIL_INSERT_PROB = 0.90

    INDEL_CONTINUATION_PROB = 0.7
    INDEL_SWITCH_PROB = 0.3

    START_AND_END_RANDOM_PROB = 0.5
    RAND_RAND_PROB = 0.5
    RAND_INS_CONTINUATION_PROB = 0.8
    RAND_INS_TO_DEL_PROB = 0.1
    RAND_INS_END_PROB = 0.1

    SUDDEN_END_PROB = 0.01
    MATCH_END_PROB = 0.1

    def __init__(self,
                 name,
                 description,
                 version,
                 array_element_structure,
                 adapters,
                 direct_connections,
                 start_element_names,
                 end_element_names,
                 do_build=True):

        self.name = name
        self.description = description
        self.version = version

        self.array_element_structure = array_element_structure
        self.adapter_dict = adapters
        self.direct_connections_dict = direct_connections

        self.start_element_names = start_element_names
        self.end_element_names = end_element_names

        self.hmm = None
        self.key_adapters = self._create_key_adapter_order()
        self.key_adapter_set = set(self.key_adapters)

        if do_build:
            self.build()

    def annotate(self, seq):
        """Annotate the given segment using this model."""
        logp, path = self.hmm.viterbi(seq)

        ppath = []
        for p, (idx, state) in enumerate(path[1:-1]):
            if (
                    "start" not in state.name
                    and ":RD" not in state.name
                    and ":D" not in state.name
            ):
                ppath.append(f'{re.split(":", state.name)[0]}')

        return logp, ppath

    def validate_segment_order(self, ordered_segment_names, allow_missing_first_adapter=True):
        """Validate the order of the given segments against the expected order in this model.

        Returns: (True|False, # key adapters found, first key adapter index)"""

        # Iterate through our given segment names and check if they occur in order:
        num_key_adapters_found = 0
        key_adapter_indx = 0
        first_key_adapter_index = 0
        found_first_key = False
        for n in ordered_segment_names:
            # Ignore all segment names that do not characterize our library:
            if n in self.key_adapter_set:

                # If this is our first segment, we should allow for the possibility that our array begins
                # somewhere after the first element.  We must find the starting point:
                if not found_first_key:
                    while n != self.key_adapters[key_adapter_indx]:
                        key_adapter_indx += 1
                    first_key_adapter_index = key_adapter_indx
                    found_first_key = True

                    # TODO: This can be eliminated for newer datasets, but is here because we're still testing with
                    #       older data that does not have the first MAS-seq overhang.  The model has already been
                    #       updated to handle these segments.
                    if allow_missing_first_adapter and (key_adapter_indx == 1) and \
                            (ordered_segment_names[0] == "10x_Adapter"):
                        num_key_adapters_found += 1
                        first_key_adapter_index = 0

                # Check our key segments here:
                # NOTE: it's possible that we can start matching in the middle of an array, but at the start of a read,
                #       and so we have to do a bounds check here:
                if key_adapter_indx < len(self.key_adapters) and (n == self.key_adapters[key_adapter_indx]):
                    key_adapter_indx += 1
                    num_key_adapters_found += 1
                else:
                    # This read does not conform to the model!
                    return False, num_key_adapters_found, first_key_adapter_index

        # If we've made it here and we have seen at least 1 key adapter, then we have a valid array:
        is_valid = True if num_key_adapters_found > 0 else False

        # Except!  If we have an array singleton and it's the terminal overhang adapter,
        # then we do NOT have a valid array:
        if is_valid and (num_key_adapters_found == 1) and \
                (self.key_adapters[first_key_adapter_index] == self.array_element_structure[-1][-1]):
            is_valid = False

        return is_valid, num_key_adapters_found, first_key_adapter_index

    def extract_key_segment_names(self, segment_names):
        """Return a list of key segment names from the given list of segment_names."""
        return [n for n in segment_names if n in self.key_adapter_set]

    def build(self):
        """Build the HMM underlying this model given our segment information."""
        self.hmm = LibraryModel._make_random_repeat_model()
        for k, v in self.adapter_dict.items():
            self.hmm.add_model(LibraryModel._make_global_alignment_model(v, k))

        self.hmm.bake(merge="None")

        # dictionary of model starting states, random start, and random end
        starts = {}
        rda = None
        rdb = None
        for s in self.hmm.states:
            if "-start" in s.name and "random" not in s.name:
                starts[re.sub("-start", "", s.name)] = s
            elif "random:RDA" in s.name:
                rda = s
            elif "random:RDB" in s.name:
                rdb = s

        # link silent start state to all of our start nodes
        # this is the start of the array.
        for sname in starts:
            if sname in self.start_element_names:
                self.hmm.add_transition(self.hmm.start, starts[sname], 1.0/len(self.start_element_names))

        # link random element model silent state A to all of our starts
        # This is the transition from `random` to one of our array elements.
        for sname in starts:
            if sname in self.end_element_names:
                self.hmm.add_transition(rda, starts[sname], 1.0/len(self.end_element_names))

        # link up adapter final states according to our direct connections dictionary
        for s in self.hmm.states:
            m = re.match(r"^(\w+):([MID])(\d+)", s.name)

            # If we're in the last state of each adapter:
            if m is not None and int(m.group(3)) == len(self.adapter_dict[m.group(1)]):
                sname = m.group(1)

                # If we have direct connections from this state to some other states then we add
                # them into the model:
                if sname in self.direct_connections_dict:
                    for dcname in self.direct_connections_dict[sname]:
                        self.hmm.add_transition(
                            s, starts[dcname], 1.0 / len(self.direct_connections_dict[sname])
                        )
                # If we have no direct connections from this state, we add in a transition from this
                # state to the silent random B state:
                else:
                    # Verify this probability is the notional equivalent:
                    self.hmm.add_transition(s, rdb, LibraryModel.RAND_RAND_PROB)

        # link up all adapters to model end state
        for s in self.hmm.states:
            m = re.match(r"^(\w+):([MID])(\d+)", s.name)

            # Get the last state in each adapter:
            if m is not None and int(m.group(3)) == len(self.adapter_dict[m.group(1)]):
                self.hmm.add_transition(s, self.hmm.end, LibraryModel.SUDDEN_END_PROB)

        self.hmm.bake()

    @staticmethod
    def create_model(model_name):
        """Create the model of the given name.
        If the name is not in LibraryModel.has_prebuilt_model, will attempt to read the given model_name as a json
        file to create a model.  This file is assumed to contain the information as per LibraryModel.to_json()."""
        # Get our model:
        if LibraryModel.has_prebuilt_model(model_name):
            logger.info(f"Using %s", LibraryModel.pre_configured_models[model_name]["description"])
            m = LibraryModel.build_pre_configured_model(model_name)
        else:
            logger.info(f"Loading model from json file: %s", model_name)
            m = LibraryModel.from_json_file(model_name)

        return m

    @staticmethod
    def has_prebuilt_model(model_name):
        """Returns True iff LibraryModel.pre_configured_models has a model with the given name.  False otherwise."""
        return model_name in LibraryModel.pre_configured_models

    def _create_key_adapter_order(self):
        """Setup an ordered list of key segments that characterize the correct array element order."""

        # TODO: Generalize this for all library types / segment names!
        # Assumption: self.array_element_structure contains the array elements in our library in the order in which
        #             they appear in the data.
        # Heuristic: The segments that characterize the array elements themselves all have single-character names, so we
        #            filter out all segments from self.array_element_structure with names longer than 1 char.  We then
        #            use these in order to characterize the reads.

        ordered_key_adapters = [s for array in self.array_element_structure
                                for s in array if len(s) == 1]
        return ordered_key_adapters

    @staticmethod
    def _make_global_alignment_model(target, name=None):
        model = HiddenMarkovModel(name=name)
        s = {}

        # add states
        i0 = State(
            DiscreteDistribution({
                    "A": LibraryModel.RANDOM_BASE_PROB,
                    "C": LibraryModel.RANDOM_BASE_PROB,
                    "G": LibraryModel.RANDOM_BASE_PROB,
                    "T": LibraryModel.RANDOM_BASE_PROB
                }),
            name=f"{name}:I0",
        )

        model.add_state(i0)

        s[i0.name] = i0

        for c in range(len(target)):
            dc = State(None, name=f"{name}:D{c + 1}")

            mc = State(
                DiscreteDistribution(
                    {
                        "A": LibraryModel.PER_BASE_MATCH_PROB if target[c] == "A" else LibraryModel.PER_BASE_MISMATCH_PROB,
                        "C": LibraryModel.PER_BASE_MATCH_PROB if target[c] == "C" else LibraryModel.PER_BASE_MISMATCH_PROB,
                        "G": LibraryModel.PER_BASE_MATCH_PROB if target[c] == "G" else LibraryModel.PER_BASE_MISMATCH_PROB,
                        "T": LibraryModel.PER_BASE_MATCH_PROB if target[c] == "T" else LibraryModel.PER_BASE_MISMATCH_PROB,
                    }
                ),
                name=f"{name}:M{c + 1}",
            )

            ic = State(
                DiscreteDistribution({
                    "A": LibraryModel.RANDOM_BASE_PROB,
                    "C": LibraryModel.RANDOM_BASE_PROB,
                    "G": LibraryModel.RANDOM_BASE_PROB,
                    "T": LibraryModel.RANDOM_BASE_PROB
                }),
                name=f"{name}:I{c + 1}",
            )

            model.add_states([mc, ic, dc])

            s[dc.name] = dc
            s[mc.name] = mc
            s[ic.name] = ic

        # add transitions
        model.add_transition(model.start, s[f"{name}:I0"], LibraryModel.MATCH_INDEL_PROB)
        model.add_transition(model.start, s[f"{name}:D1"], LibraryModel.MATCH_INDEL_PROB)
        model.add_transition(model.start, s[f"{name}:M1"], LibraryModel.MATCH_MATCH_PROB)

        model.add_transition(s[f"{name}:I0"], s[f"{name}:I0"], LibraryModel.INDEL_CONTINUATION_PROB)
        model.add_transition(s[f"{name}:I0"], s[f"{name}:D1"], LibraryModel.INDEL_SWITCH_PROB/2)
        model.add_transition(s[f"{name}:I0"], s[f"{name}:M1"], LibraryModel.INDEL_SWITCH_PROB/2)

        for c in range(1, len(target)):
            model.add_transition(s[f"{name}:D{c}"], s[f"{name}:D{c + 1}"], LibraryModel.INDEL_SWITCH_PROB/2)
            model.add_transition(s[f"{name}:D{c}"], s[f"{name}:I{c}"], LibraryModel.INDEL_CONTINUATION_PROB)
            model.add_transition(s[f"{name}:D{c}"], s[f"{name}:M{c + 1}"], LibraryModel.INDEL_SWITCH_PROB/2)

            model.add_transition(s[f"{name}:I{c}"], s[f"{name}:D{c + 1}"], LibraryModel.INDEL_SWITCH_PROB/2)
            model.add_transition(s[f"{name}:I{c}"], s[f"{name}:I{c}"], LibraryModel.INDEL_SWITCH_PROB/2)
            model.add_transition(s[f"{name}:I{c}"], s[f"{name}:M{c + 1}"], LibraryModel.INDEL_CONTINUATION_PROB)

            model.add_transition(s[f"{name}:M{c}"], s[f"{name}:D{c + 1}"], LibraryModel.MATCH_INDEL_PROB)
            model.add_transition(s[f"{name}:M{c}"], s[f"{name}:I{c}"], LibraryModel.MATCH_INDEL_PROB)
            model.add_transition(s[f"{name}:M{c}"], s[f"{name}:M{c + 1}"], LibraryModel.MATCH_MATCH_PROB)

        model.add_transition(
            s[f"{name}:D{len(target)}"], s[f"{name}:I{len(target)}"], LibraryModel.INDEL_CONTINUATION_PROB
        )
        model.add_transition(s[f"{name}:D{len(target)}"], model.end, LibraryModel.INDEL_SWITCH_PROB)

        model.add_transition(s[f"{name}:I{len(target)}"], s[f"{name}:I{len(target)}"], LibraryModel.INDEL_SWITCH_PROB/2)
        model.add_transition(
            s[f"{name}:I{len(target)}"],
            model.end,
            LibraryModel.INDEL_CONTINUATION_PROB + (LibraryModel.INDEL_SWITCH_PROB/2)
        )

        model.add_transition(
            s[f"{name}:M{len(target)}"], s[f"{name}:I{len(target)}"], LibraryModel.MATCH_TRAIL_INSERT_PROB
        )
        model.add_transition(s[f"{name}:M{len(target)}"], model.end, LibraryModel.MATCH_END_PROB)

        model.bake(merge="None")

        return model

    @staticmethod
    def _make_random_repeat_model(name="random"):
        model = HiddenMarkovModel(name=name)

        # add states
        ri = State(
            DiscreteDistribution({
                "A": LibraryModel.RANDOM_BASE_PROB,
                "C": LibraryModel.RANDOM_BASE_PROB,
                "G": LibraryModel.RANDOM_BASE_PROB,
                "T": LibraryModel.RANDOM_BASE_PROB
            }),
            name=f"{name}:RI",
        )
        rda = State(None, name=f"{name}:RDA")
        rdb = State(None, name=f"{name}:RDB")

        model.add_states([ri, rda, rdb])

        # add transitions
        model.add_transition(model.start, rda, LibraryModel.START_AND_END_RANDOM_PROB)
        model.add_transition(model.start, ri, LibraryModel.START_AND_END_RANDOM_PROB)

        model.add_transition(ri, ri, LibraryModel.RAND_INS_CONTINUATION_PROB)
        model.add_transition(ri, rda, LibraryModel.RAND_INS_TO_DEL_PROB)
        model.add_transition(ri, model.end, LibraryModel.RAND_INS_END_PROB)

        model.add_transition(rdb, ri, LibraryModel.RAND_RAND_PROB)
        model.add_transition(rdb, model.end, LibraryModel.START_AND_END_RANDOM_PROB)

        model.bake(merge="None")

        return model

    # @staticmethod
    # def _make_homopolymer_repeat_model(name="homopolymer_repeat", base="A"):
    #     model = HiddenMarkovModel(name=name)
    #
    #     # add states
    #     hpr = State(
    #         DiscreteDistribution({
    #             base: 1
    #         }),
    #         name=f"{name}:RI",
    #     )
    #     hpr_a = State(None, name=f"{name}:HPRA")
    #     hpr_b = State(None, name=f"{name}:HPRB")
    #
    #     model.add_states([hpr, hpr_a, hpr_b])
    #
    #     # add transitions
    #     model.add_transition(model.start, hpr_a, LibraryModel.START_AND_END_RANDOM_PROB)
    #     model.add_transition(model.start, hpr, LibraryModel.START_AND_END_RANDOM_PROB)
    #
    #     model.add_transition(hpr, hpr, LibraryModel.RAND_INS_CONTINUATION_PROB)
    #     model.add_transition(hpr, hpr_a, LibraryModel.RAND_INS_TO_DEL_PROB)
    #     model.add_transition(hpr, model.end, LibraryModel.RAND_INS_END_PROB)
    #
    #     model.add_transition(hpr_b, hpr, LibraryModel.RAND_RAND_PROB)
    #     model.add_transition(hpr_b, model.end, LibraryModel.START_AND_END_RANDOM_PROB)
    #
    #     model.bake(merge="None")
    #
    #     return model

    def to_json(self, outfile=None, indent=4):
        """Serialize this model to a json object and return that json object.
        If outfile is not none, will write the json object to the given file path."""

        model_data = {
            "name": self.name,
            "description": self.description,
            "version": self.version,
            "array_element_structure": self.array_element_structure,
            "adapters": self.adapter_dict,
            "direct_connections": {k: list(v) for k, v in self.direct_connections_dict.items()},
            "start_element_names": list(self.start_element_names),
            "end_element_names": list(self.end_element_names)
        }

        if outfile:
            with open(outfile, 'w') as f:
                json.dump(model_data, f, indent=indent)

        return json.dumps(model_data, indent=indent)

    @staticmethod
    def from_json_file(json_file):
        """Create a LibraryModel instance from the given json data.
        This method will open the file at the given location and use the data in that file to create a LibraryModel."""

        try:
            with open(json_file) as f:
                json_data = json.load(f)
        except FileNotFoundError:
            logger.error(f"File does not exist: {json_file}")
            sys.exit(1)

        m = LibraryModel(
            name=json_data["name"],
            description=json_data["description"],
            version=json_data["version"],
            array_element_structure=tuple(tuple(v) for v in json_data["array_element_structure"]),
            adapters=json_data["adapters"],
            direct_connections={k: set(v) for k, v in json_data["direct_connections"].items()},
            start_element_names=set(json_data["start_element_names"]),
            end_element_names=set(json_data["end_element_names"]),
        )

        return m

    @staticmethod
    def build_pre_configured_model(model_name):
        """Build a pre-configured model based on the given model name.
        If the given name does not appear in self.pre_configured_models, then a KeyError will be thrown."""
        if model_name not in LibraryModel.pre_configured_models:
            raise KeyError(f"Model not found in pre-configured models: {model_name}")

        return LibraryModel(
            name=model_name,
            description=LibraryModel.pre_configured_models[model_name]["description"],
            version=LibraryModel.pre_configured_models[model_name]["version"],
            array_element_structure=LibraryModel.pre_configured_models[model_name]["array_element_structure"],
            adapters=LibraryModel.pre_configured_models[model_name]["adapters"],
            direct_connections=LibraryModel.pre_configured_models[model_name]["direct_connections"],
            start_element_names=LibraryModel.pre_configured_models[model_name]["start_element_names"],
            end_element_names=LibraryModel.pre_configured_models[model_name]["end_element_names"],
        )

    # TODO: Make an enum for this...
    pre_configured_models = {
        "mas15": {
            "description": "The standard MAS-seq 15 array element model.",
            "version": "1.0.0",
            "array_element_structure": (
                # NOTE: the first element doesn't currently have the "A" adapter in this version of the library.
                ("A", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("B", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("C", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("D", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("E", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("F", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("G", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("H", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("I", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("J", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("K", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("L", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("M", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("N", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                # The last element doesn't currently have the "P" adapter in this version of the library:
                ("O", "10x_Adapter", "random", "Poly_A", "3p_Adapter", "P"),
            ),
            "adapters": {
                "10x_Adapter": "TCTACACGACGCTCTTCCGATCT",
                "Poly_A": "A" * 30,
                "3p_Adapter": "GTACTCTGCGTTGATACCACTGCTT",
                "A": "AGCTTACTTGTGAAGA",
                "B": "ACTTGTAAGCTGTCTA",
                "C": "ACTCTGTCAGGTCCGA",
                "D": "ACCTCCTCCTCCAGAA",
                "E": "AACCGGACACACTTAG",
                "F": "AGAGTCCAATTCGCAG",
                "G": "AATCAAGGCTTAACGG",
                "H": "ATGTTGAATCCTAGCG",
                "I": "AGTGCGTTGCGAATTG",
                "J": "AATTGCGTAGTTGGCC",
                "K": "ACACTTGGTCGCAATC",
                "L": "AGTAAGCCTTCGTGTC",
                "M": "ACCTAGATCAGAGCCT",
                "N": "AGGTATGCCGGTTAAG",
                "O": "AAGTCACCGGCACCTT",
                "P": "ATGAAGTGGCTCGAGA",
            },
            "direct_connections": {
                "Poly_A": {"3p_Adapter"},
                "3p_Adapter": {
                    "A",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                    "G",
                    "H",
                    "I",
                    "J",
                    "K",
                    "L",
                    "M",
                    "N",
                    "O",
                    "P",
                },
                "A": {"10x_Adapter"},
                "B": {"10x_Adapter"},
                "C": {"10x_Adapter"},
                "D": {"10x_Adapter"},
                "E": {"10x_Adapter"},
                "F": {"10x_Adapter"},
                "G": {"10x_Adapter"},
                "H": {"10x_Adapter"},
                "I": {"10x_Adapter"},
                "J": {"10x_Adapter"},
                "K": {"10x_Adapter"},
                "L": {"10x_Adapter"},
                "M": {"10x_Adapter"},
                "N": {"10x_Adapter"},
                "O": {"10x_Adapter"},
                "P": {"10x_Adapter"},
            },
            "start_element_names": {"A", "10x_Adapter"},
            "end_element_names": {"Poly_A", "P"},
        },
        "mas15threeP": {
            "description": "The 3' kit MAS-seq 15 array element model.",
            "version": "1.0.0",
            "array_element_structure": (
                # NOTE: the first element doesn't currently have the "A" adapter in this version of the library.
                ("A", "10x_Adapter", "random", "TSO"),
                ("B", "10x_Adapter", "random", "TSO"),
                ("C", "10x_Adapter", "random", "TSO"),
                ("D", "10x_Adapter", "random", "TSO"),
                ("E", "10x_Adapter", "random", "TSO"),
                ("F", "10x_Adapter", "random", "TSO"),
                ("G", "10x_Adapter", "random", "TSO"),
                ("H", "10x_Adapter", "random", "TSO"),
                ("I", "10x_Adapter", "random", "TSO"),
                ("J", "10x_Adapter", "random", "TSO"),
                ("K", "10x_Adapter", "random", "TSO"),
                ("L", "10x_Adapter", "random", "TSO"),
                ("M", "10x_Adapter", "random", "TSO"),
                ("N", "10x_Adapter", "random", "TSO"),
                # The last element doesn't currently have the "P" adapter in this version of the library:
                ("O", "10x_Adapter", "random", "TSO", "P"),
            ),
            "adapters": {
                "10x_Adapter": "TCTACACGACGCTCTTCCGATCT",
                "TSO": "CCCATGTACTCTGCGTTGATACCACTGCTT",
                "A": "AGCTTACTTGTGAAGA",
                "B": "ACTTGTAAGCTGTCTA",
                "C": "ACTCTGTCAGGTCCGA",
                "D": "ACCTCCTCCTCCAGAA",
                "E": "AACCGGACACACTTAG",
                "F": "AGAGTCCAATTCGCAG",
                "G": "AATCAAGGCTTAACGG",
                "H": "ATGTTGAATCCTAGCG",
                "I": "AGTGCGTTGCGAATTG",
                "J": "AATTGCGTAGTTGGCC",
                "K": "ACACTTGGTCGCAATC",
                "L": "AGTAAGCCTTCGTGTC",
                "M": "ACCTAGATCAGAGCCT",
                "N": "AGGTATGCCGGTTAAG",
                "O": "AAGTCACCGGCACCTT",
                "P": "ATGAAGTGGCTCGAGA",
            },
            "direct_connections": {
                "TSO": {
                    "A",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                    "G",
                    "H",
                    "I",
                    "J",
                    "K",
                    "L",
                    "M",
                    "N",
                    "O",
                    "P",
                },
                "A": {"10x_Adapter"},
                "B": {"10x_Adapter"},
                "C": {"10x_Adapter"},
                "D": {"10x_Adapter"},
                "E": {"10x_Adapter"},
                "F": {"10x_Adapter"},
                "G": {"10x_Adapter"},
                "H": {"10x_Adapter"},
                "I": {"10x_Adapter"},
                "J": {"10x_Adapter"},
                "K": {"10x_Adapter"},
                "L": {"10x_Adapter"},
                "M": {"10x_Adapter"},
                "N": {"10x_Adapter"},
                "O": {"10x_Adapter"},
                "P": {"10x_Adapter"},
            },
            "start_element_names": {"A", "10x_Adapter"},
            "end_element_names": {"P"},
        },
        "mas10": {
            "description": "The MAS-seq 10 array element model.",
            "version": "1.0.0",
            "array_element_structure": (
                ("Q", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("C", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("M", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("I", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("O", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("J", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("B", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("D", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                ("K", "10x_Adapter", "random", "Poly_A", "3p_Adapter"),
                # The last element may not currently have the "R" adapter in this version of the library:
                ("H", "10x_Adapter", "random", "Poly_A", "3p_Adapter", "R"),
            ),
            "adapters": {
                "10x_Adapter": "TCTACACGACGCTCTTCCGATCT",
                "Poly_A": "A" * 30,
                "3p_Adapter": "GTACTCTGCGTTGATACCACTGCTT",
                "B": "ACTTGTAAGCTGTCTA",
                "C": "ACTCTGTCAGGTCCGA",
                "D": "ACCTCCTCCTCCAGAA",
                "H": "ATGTTGAATCCTAGCG",
                "I": "AGTGCGTTGCGAATTG",
                "J": "AATTGCGTAGTTGGCC",
                "K": "ACACTTGGTCGCAATC",
                "M": "ACCTAGATCAGAGCCT",
                "O": "AAGTCACCGGCACCTT",
                "Q": "AAGCACCATAATGTGT",
                "R": "AACCGGACACACTTAG",
            },
            "direct_connections": {
                "Poly_A": {"3p_Adapter"},
                "3p_Adapter": {
                    "Q",
                    "C",
                    "M",
                    "I",
                    "O",
                    "J",
                    "B",
                    "D",
                    "K",
                    "H",
                    "R",
                },
                "B": {"10x_Adapter"},
                "C": {"10x_Adapter"},
                "D": {"10x_Adapter"},
                "H": {"10x_Adapter"},
                "I": {"10x_Adapter"},
                "J": {"10x_Adapter"},
                "K": {"10x_Adapter"},
                "M": {"10x_Adapter"},
                "O": {"10x_Adapter"},
                "Q": {"10x_Adapter"},
                "R": {"10x_Adapter"},
            },
            "start_element_names": {"Q", "10x_Adapter"},
            "end_element_names": {"Poly_A", "R"},
        },
        "mas10threeP": {
            "description": "The 3' kit MAS-seq 10 array element model.",
            "version": "1.0.0",
            "array_element_structure": (
                ("Q", "10x_Adapter", "random", "TSO"),
                ("C", "10x_Adapter", "random", "TSO"),
                ("M", "10x_Adapter", "random", "TSO"),
                ("I", "10x_Adapter", "random", "TSO"),
                ("O", "10x_Adapter", "random", "TSO"),
                ("J", "10x_Adapter", "random", "TSO"),
                ("B", "10x_Adapter", "random", "TSO"),
                ("D", "10x_Adapter", "random", "TSO"),
                ("K", "10x_Adapter", "random", "TSO"),
                # The last element may not currently have the "R" adapter in this version of the library:
                ("H", "10x_Adapter", "random", "TSO", "R"),
            ),
            "adapters": {
                "10x_Adapter": "TCTACACGACGCTCTTCCGATCT",
                "TSO": "CCCATGTACTCTGCGTTGATACCACTGCTT",
                "B": "ACTTGTAAGCTGTCTA",
                "C": "ACTCTGTCAGGTCCGA",
                "D": "ACCTCCTCCTCCAGAA",
                "H": "ATGTTGAATCCTAGCG",
                "I": "AGTGCGTTGCGAATTG",
                "J": "AATTGCGTAGTTGGCC",
                "K": "ACACTTGGTCGCAATC",
                "M": "ACCTAGATCAGAGCCT",
                "O": "AAGTCACCGGCACCTT",
                "Q": "AAGCACCATAATGTGT",
                "R": "AACCGGACACACTTAG",
            },
            "direct_connections": {
                "TSO": {
                    "Q",
                    "C",
                    "M",
                    "I",
                    "O",
                    "J",
                    "B",
                    "D",
                    "K",
                    "H",
                    "R",
                },
                "B": {"10x_Adapter"},
                "C": {"10x_Adapter"},
                "D": {"10x_Adapter"},
                "H": {"10x_Adapter"},
                "I": {"10x_Adapter"},
                "J": {"10x_Adapter"},
                "K": {"10x_Adapter"},
                "M": {"10x_Adapter"},
                "O": {"10x_Adapter"},
                "Q": {"10x_Adapter"},
                "R": {"10x_Adapter"},
            },
            "start_element_names": {"Q", "10x_Adapter"},
            "end_element_names": {"R"},
        },
        "slide-seq": {
            # The slide-seq model is:
            #
            #                 |-----10x_Adapter---->        |--splitter------>               |------Poly_T---------------->                  |--------5p_Adapter----------|                         # noqa
            # AGCTTACTTGTGAAGACTACACGACGCTCTTCCGATCTNNNNNNNNTCTTCAGCGTTCCCGAGANNNNNNNNNNNNNVVTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTVNNNNNNNNNNNNNNNNNCCCATGTACTCTGCGTTGATACCACTGCTTACTTGTAAGCTGTCTA...      # noqa
            # |------A------->                      <------|                  <-----------|                                 <----cDNA-------|                              |-------B------>         # noqa
            #                                          V                           V
            #                                    Spatial Barcode 2         Spatial Barcode 1
            "description": "The Slide-seq 15 array element model.",
            "version": "0.0.1",
            "array_element_structure": (
                # Currently longbow can't handle multiple `random` segments, so we'll need to
                # update this in the future to look more like this:
                # ("A", "10x_Adapter", "random", "barcode_splitter", "random", "Poly_T", "random", "5p_Adapter"),
                ("A", "10x_Adapter", "random", "5p_Adapter"),
                ("B", "10x_Adapter", "random", "5p_Adapter"),
                ("C", "10x_Adapter", "random", "5p_Adapter"),
                ("D", "10x_Adapter", "random", "5p_Adapter"),
                ("E", "10x_Adapter", "random", "5p_Adapter"),
                ("F", "10x_Adapter", "random", "5p_Adapter"),
                ("G", "10x_Adapter", "random", "5p_Adapter"),
                ("H", "10x_Adapter", "random", "5p_Adapter"),
                ("I", "10x_Adapter", "random", "5p_Adapter"),
                ("J", "10x_Adapter", "random", "5p_Adapter"),
                ("K", "10x_Adapter", "random", "5p_Adapter"),
                ("L", "10x_Adapter", "random", "5p_Adapter"),
                ("M", "10x_Adapter", "random", "5p_Adapter"),
                ("N", "10x_Adapter", "random", "5p_Adapter"),
                ("O", "10x_Adapter", "random", "5p_Adapter", "P"),
            ),
            "adapters": {
                "10x_Adapter": "TCTACACGACGCTCTTCCGATCT",
                "5p_Adapter": "CCCATGTACTCTGCGTTGATACCACTGCTT",
                "A": "AGCTTACTTGTGAAGA",
                "B": "ACTTGTAAGCTGTCTA",
                "C": "ACTCTGTCAGGTCCGA",
                "D": "ACCTCCTCCTCCAGAA",
                "E": "AACCGGACACACTTAG",
                "F": "AGAGTCCAATTCGCAG",
                "G": "AATCAAGGCTTAACGG",
                "H": "ATGTTGAATCCTAGCG",
                "I": "AGTGCGTTGCGAATTG",
                "J": "AATTGCGTAGTTGGCC",
                "K": "ACACTTGGTCGCAATC",
                "L": "AGTAAGCCTTCGTGTC",
                "M": "ACCTAGATCAGAGCCT",
                "N": "AGGTATGCCGGTTAAG",
                "O": "AAGTCACCGGCACCTT",
                "P": "ATGAAGTGGCTCGAGA",
            },
            "direct_connections": {
                "5p_Adapter": {
                    "A",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                    "G",
                    "H",
                    "I",
                    "J",
                    "K",
                    "L",
                    "M",
                    "N",
                    "O",
                    "P",
                },
                "A": {"10x_Adapter"},
                "B": {"10x_Adapter"},
                "C": {"10x_Adapter"},
                "D": {"10x_Adapter"},
                "E": {"10x_Adapter"},
                "F": {"10x_Adapter"},
                "G": {"10x_Adapter"},
                "H": {"10x_Adapter"},
                "I": {"10x_Adapter"},
                "J": {"10x_Adapter"},
                "K": {"10x_Adapter"},
                "L": {"10x_Adapter"},
                "M": {"10x_Adapter"},
                "N": {"10x_Adapter"},
                "O": {"10x_Adapter"}
            },
            # Right now it won't work properly without both A and 10xAdapter being starts and P and
            # 5pAdapter being ends.
            "start_element_names": {"A", "10x_Adapter"},
            "end_element_names": {"P", "5p_Adapter"},
        },
        "mas8prototype": {
            "description": "The prototype MAS-seq 8 array element model.",
            "version": "1.0.0",
            "array_element_structure": (
                # NOTE: the first element may not have the "A" adapter in this version of the library.
                ("A", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("B", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("C", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("D", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("E", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("F", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("G", "10x_Adapter", "random", "Poly_T", "random", "TSO"),
                ("H", "10x_Adapter", "random", "Poly_T", "random", "TSO", "A"),
            ),
            "adapters": {
                "10x_Adapter": "CTACACGACGCTCTTCCGATCT",
                "Poly_T": "T" * 30,
                "TSO": "CCCATGTACTCTGCGTTGATACCACTGCTT",
                "A": "ACGTACAGT",
                "B": "AAACTGCAT",
                "C": "AGAGTCACT",
                "D": "AGCAAGTTT",
                "E": "AAGTGGTGT",
                "F": "AGTGGACTT",
                "G": "ACAGGTTAT",
                "H": "ATCTCACAT",
            },
            "direct_connections": {
                "TSO": {
                    "A",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                    "G",
                    "H",
                },
                "A": {"10x_Adapter"},
                "B": {"10x_Adapter"},
                "C": {"10x_Adapter"},
                "D": {"10x_Adapter"},
                "E": {"10x_Adapter"},
                "F": {"10x_Adapter"},
                "G": {"10x_Adapter"},
                "H": {"10x_Adapter"},
            },
            "start_element_names": {"A", "10x_Adapter"},
            "end_element_names": {"TSO", "A"},
        }
    }


# IUPAC RC's from: http://arep.med.harvard.edu/labgc/adnan/projects/Utilities/revcomp.html
# and https://www.dnabaser.com/articles/IUPAC%20ambiguity%20codes.html
RC_BASE_MAP = {
    "N": "N",
    "A": "T",
    "T": "A",
    "G": "C",
    "C": "G",
    "Y": "R",
    "R": "Y",
    "S": "S",
    "W": "W",
    "K": "M",
    "M": "K",
    "B": "V",
    "V": "B",
    "D": "H",
    "H": "D",
    "n": "n",
    "a": "t",
    "t": "a",
    "g": "c",
    "c": "g",
    "y": "r",
    "r": "y",
    "s": "s",
    "w": "w",
    "k": "m",
    "m": "k",
    "b": "v",
    "v": "b",
    "d": "h",
    "h": "d",
}


def reverse_complement(base_string):
    """
    Reverse complements the given base_string.
    :param base_string: String of bases to be reverse-complemented.
    :return: The reverse complement of the given base string.
    """

    return "".join(map(lambda b: RC_BASE_MAP[b], base_string[::-1]))



import logging
import time
import sys
import os
import re
import math
import ssw

import click
import click_log

import gzip
import pysam
from collections import OrderedDict
from construct import *

import matplotlib.pyplot as plt
from matplotlib import transforms

from ..utils.model import LibraryModel
from ..utils.model import reverse_complement
from ..utils import bam_utils

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("inspect")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-r",
    "--read-names",
    type=str,
    multiple=True,
    help="read names (or file(s) of read names) to inspect",
)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(exists=True),
    help="BAM .pbi index file",
)
@click.option(
    "-f",
    "--file-format",
    default="png",
    type=click.Choice(["png", "pdf"]),
    help="Image file format",
)
@click.option(
    "-o",
    "--outdir",
    default=".",
    required=False,
    type=click.Path(exists=False),
    help="Output directory",
)
@click.option(
    "-m",
    "--model",
    default="mas15",
    show_default=True,
    help="The model to use for annotation.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name and Longbow will attempt to "
         "read in the file and create a LibraryModel from it.  Longbow will assume the contents are the configuration "
         "of a LibraryModel as per LibraryModel.to_json()."
)
@click.option(
    '--seg-score',
    is_flag=True,
    default=False,
    show_default=True,
    help="Display alignment score for annotated segments."
)
@click.option(
    "--max-length",
    type=int,
    default=60000,
    show_default=True,
    required=False,
    help="Maximum length of a read to process.  Reads beyond this length will not be annotated.  If the input file has already been annotated, this parameter is ignored."
)
@click.option(
    "--min-rq",
    type=float,
    default=-2,
    show_default=True,
    required=False,
    help="Minimum ccs-determined read quality for a read to be annotated.  CCS read quality range is [-1,1].  If the input file has already been annotated, this parameter is ignored."
)
@click.argument("input-bam", type=click.Path(exists=True))
def main(read_names, pbi, file_format, outdir, model, seg_score, max_length, min_rq, input_bam):
    """Inspect the classification results on specified reads."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    # Get our model:
    if LibraryModel.has_prebuilt_model(model):
        logger.info(f"Using %s", LibraryModel.pre_configured_models[model]["description"])
        lb_model = LibraryModel.build_pre_configured_model(model)
    else:
        logger.info(f"Loading model from json file: %s", model)
        lb_model = LibraryModel.from_json_file(model)

    # Create an aligner if we are scoring our segments:
    ssw_aligner = ssw.Aligner() if seg_score else None

    # Open our bam file:
    pysam.set_verbosity(0)
    with pysam.AlignmentFile(input_bam, "rb", check_sq=False, require_index=False) as bf:
        # If we have read names, we should use them to inspect the file:
        if len(read_names) > 0:

            pbi = f"{input_bam}.pbi" if pbi is None else pbi
            if not os.path.exists(pbi):
                raise FileNotFoundError(f"Missing .pbi file for {input_bam}")

            file_offsets = load_read_offsets(pbi, load_read_names(read_names))

            for i, z in enumerate(file_offsets):

                if not file_offsets[z]["offset"]:
                    logger.error("Read not in index file: %s", read_names[i])
                    sys.exit(1)

                bf.seek(file_offsets[z]["offset"])
                read = bf.__next__()
                __create_read_figure(file_format, lb_model, outdir, read, seg_score, ssw_aligner, max_length, min_rq)
        else:
            # Without read names we just inspect every read in the file:
            logger.info("No read names given.  Inspecting every read in the input bam file.")
            for read in bf:
                __create_read_figure(file_format, lb_model, outdir, read, seg_score, ssw_aligner, max_length, min_rq)

    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)


def __create_read_figure(file_format, lb_model, outdir, read, seg_score, ssw_aligner, max_length, min_rq):
    """Create a figure for the given read."""

    out = f'{outdir}/{re.sub("/", "_", read.query_name)}.{file_format}'

    seq, path, logp = annotate_read(read, lb_model, max_length, min_rq)

    if seq is not None:
        logger.info("Drawing read '%s' to '%s'", read.query_name, out)
        draw_state_sequence(seq, path, logp, read, out, seg_score, lb_model, ssw_aligner, size=13, family="monospace")


def load_read_names(read_name_args):
    read_names = []

    for r in read_name_args:
        if os.path.exists(r):
            with open(r, "r") as f:
                for i, line in enumerate(f):
                    # Space / Tab / Newline / Line feed are all forbidden in read names by the sam spec, so we can
                    # trim it all off:
                    rn = line.strip()
                    if not bam_utils.PB_READ_NAME_RE.match(rn):
                        logger.error(
                            "Read name on line %d of file %s doesn't appear to be a PacBio read: %s", i, r, rn
                        )
                        sys.exit(1)

                    read_names.append(rn)
        else:
            if not bam_utils.PB_READ_NAME_RE.match(r):
                logger.error("Read name doesn't appear to be a PacBio read: %s", r)
                sys.exit(1)
            read_names.append(r)

    return read_names


def load_read_offsets(pbi_file, read_names):
    """
    Compute file offsets for specified read names
    """

    # Decode PacBio .pbi file.  This is not a full decode of the index, only the parts we need for read selection.
    # More on index format at https://pacbiofileformats.readthedocs.io/en/9.0/PacBioBamIndex.html .

    fmt = Struct(
        # Header
        "magic" / Const(b"PBI\x01"),
        "version_patch" / Int8ul,
        "version_minor" / Int8ul,
        "version_major" / Int8ul,
        "version_empty" / Int8ul,
        "pbi_flags" / Int16ul,
        "n_reads" / Int32ul,
        "reserved" / Padding(18),
        # Basic information section (columnar format)
        "rgId" / Padding(this.n_reads * 4),
        "qStart" / Padding(this.n_reads * 4),
        "qEnd" / Array(this.n_reads, Int32sl),
        "holeNumber" / Array(this.n_reads, Int32sl),
        "readQual" / Padding(this.n_reads * 4),
        "ctxtFlag" / Padding(this.n_reads * 1),
        "fileOffset" / Array(this.n_reads, Int64sl),
    )

    # Make a list of bgzf virtual file offsets for sharding and store ZMW counts.
    file_offsets_hash = OrderedDict()
    for read_name in read_names:
        name_pieces = re.split("/", read_name)

        hole_number = int(name_pieces[1])
        if "ccs" in name_pieces:
            if len(name_pieces) > 3:
                rr = re.split("_", name_pieces[3])
                range_end = int(rr[1]) - int(rr[0]) + 1
            else:
                range_end = -1
        else:
            range_end = re.split("_", name_pieces[2])[1]

        file_offsets_hash[f"{hole_number}_{range_end}"] = {"read_name": read_name, "offset": None}

    with gzip.open(pbi_file, "rb") as f:
        idx_contents = fmt.parse_stream(f)

        for j in range(0, idx_contents.n_reads):
            key1 = f"{idx_contents.holeNumber[j]}_{idx_contents.qEnd[j]}"
            key2 = f"{idx_contents.holeNumber[j]}_-1"

            # Save virtual file offsets for the selected reads only
            if key1 in file_offsets_hash:
                file_offsets_hash[key1]["offset"] = idx_contents.fileOffset[j]
            elif key2 in file_offsets_hash:
                file_offsets_hash[key2]["offset"] = idx_contents.fileOffset[j]

    return file_offsets_hash


def annotate_read(read, m, max_length, min_rq):
    flogp = -math.inf
    fseq = read.query_sequence
    fppath = []

    if read.has_tag("SG"):
        tag = re.split(",", read.get_tag("SG"))

        for e in tag:
            state, rrange = re.split(":", e)
            qStart, qEnd = re.split("-", rrange)
            qLen = int(qEnd) - int(qStart) + 1

            fppath.extend([state] * qLen)

        # Set our logp from the bam file:
        flogp = read.get_tag(bam_utils.READ_MODEL_SCORE_TAG)
    else:

        # Check for max length and min quality:
        if len(read.query_sequence) > max_length:
            logger.warning(f"Read is longer than max length.  "
                           f"Skipping: {read.query_name} ({len(read.query_sequence)} > {max_length})")
            return [None] * 3
        elif read.get_tag("rq") < min_rq:
            logger.warning(f"Read quality is below the minimum.  "
                           f"Skipping: {read.query_name} ({read.get_tag('rq')} < {min_rq})")
            return [None] * 3
        for seq in [read.query_sequence, reverse_complement(read.query_sequence)]:
            logp, ppath = m.annotate(seq)

            if logp > flogp:
                fseq = seq
                flogp = logp
                fppath = ppath

    return fseq, fppath, flogp


def format_state_sequence(seq, path, line_length=150):
    # TODO: Must tie this into the model itself.  We shouldn't re-define our segments here.
    adapter_state_color = "#DF5A49"
    scaffold_state_color = "#334D5C"
    poly_a_color = "#45B29D"
    three_p_adapter_color = "#EFC94C"
    random_color = "#aaaaaa"

    # Color for states we haven't enumerated:
    default_color = "#c2a8f0"

    color_hash = {
        "10x_Adapter": scaffold_state_color,
        "5p_TSO": scaffold_state_color,
        "Poly_A": poly_a_color,
        "3p_Adapter": three_p_adapter_color,
        "A": adapter_state_color,
        "B": adapter_state_color,
        "C": adapter_state_color,
        "D": adapter_state_color,
        "E": adapter_state_color,
        "F": adapter_state_color,
        "G": adapter_state_color,
        "H": adapter_state_color,
        "I": adapter_state_color,
        "J": adapter_state_color,
        "K": adapter_state_color,
        "L": adapter_state_color,
        "M": adapter_state_color,
        "N": adapter_state_color,
        "O": adapter_state_color,
        "P": adapter_state_color,
        "Q": adapter_state_color,
        "R": adapter_state_color,
        "random": random_color,
    }

    labelled_bases = []
    state_labels = []
    state_colors = []

    for i in range(0, len(path), line_length):
        bases = []
        label = path[i]

        for j in range(i, min(i + line_length, len(path))):
            a = seq[j]
            b = path[j]

            # Handle the line breaks:
            if label == b:
                bases.append(a)
            elif label != b:
                try:
                    color = color_hash[label]
                except KeyError:
                    color = default_color

                labelled_bases.append("".join(bases))
                state_labels.append(label)
                state_colors.append(color)

                bases = [a]
                label = b

        labelled_bases.append("".join(bases))
        state_labels.append(label)
        try:
            color = color_hash[label]
        except KeyError:
            color = default_color
        state_colors.append(color)

    return labelled_bases, state_colors, state_labels


def draw_state_sequence(seq, path, logp, read, out, show_seg_score, model, ssw_aligner, **kwargs):

    line_length = 150

    base_strings, colors, labels = format_state_sequence(seq, path, line_length=line_length)

    f = plt.figure(figsize=(24, 24))

    ax = plt.gca()
    t = ax.transData
    canvas = ax.figure.canvas

    qual_string = f"[read qual: {read.get_tag('rq'):0.03f}/1.000]    " if read.has_tag("rq") else ""
    np_string = f"[# Passes: {read.get_tag('np')}]    " if read.has_tag("np") else ""

    f.suptitle(
        f"{read.query_name}\n{qual_string}{np_string}[{len(read.query_sequence)} bp]    [{model.name} model score: {logp:.2f}]",
        fontsize=16
    )

    f.patch.set_visible(False)
    ax.axis("off")

    columns = line_length
    rows = 4 * 80

    plt.xlim([0, columns])
    plt.ylim([0, rows])

    letters_seen = 0
    row = 0
    column = 0
    n = 0

    for base_string, color, lbl in zip(base_strings, colors, labels):
        if column == 0:
            ytic = ax.text(0, rows - row, f"{n}  ", transform=t, ha="right")

        text = ax.text(
            0,
            rows - row,
            base_string,
            color=color,
            transform=t,
            bbox=dict(facecolor=f"{color}22", edgecolor=f"{color}22", pad=0),
            **kwargs,
        )

        # Write classified sequence
        text.draw(canvas.get_renderer())
        ex = text.get_window_extent()

        # I truly have no idea why I need the 0.72 scaling factor, but if I don't have this, PDF images
        # are super broken.
        scaling_factor = 1.00 if out.endswith(".png") else 0.72
        t = transforms.offset_copy(
            text.get_transform(), x=scaling_factor * ex.width, units="dots"
        )

        # Write state label
        if lbl != "random":

            # If we want to show the segment scores, we calculate them here:
            if show_seg_score:
                known_segment_seq = model.adapter_dict[lbl]
                alignment = ssw_aligner.align(base_string.upper(), known_segment_seq)
                optimal_score = alignment.score
                # The max score is the match score * the length of the reference segment
                max_score = len(known_segment_seq) * ssw_aligner.matrix.get_match()
                seg_score_string = f" ({optimal_score}/{max_score})"
            else:
                seg_score_string = ""

            ax.text(
                0 - (len(base_string) / 2),
                rows - row - 3.5,
                f"{lbl}{seg_score_string}",
                transform=t,
                va="bottom",
                ha="center",
                fontsize=8,
                bbox=dict(facecolor="white", edgecolor="black"),
            )

        if lbl == "10x_Adapter":
            for o in range(10, 50, 10):
                if letters_seen + len(base_string) + o < line_length:
                    ax.text(0, rows - row + 1.7, f"{' ' * o}'", transform=t, **kwargs)

        # Decide whether we need to break into a new row
        letters_seen += len(base_string)
        n += len(base_string)
        column += 1

        if letters_seen >= columns:
            letters_seen = 0

            row += 1
            column = 0

            text = ax.text(column, row, "")
            text.draw(canvas.get_renderer())
            t = transforms.offset_copy(
                text.get_transform(), x=0, y=-30 * row, units="dots"
            )

    y2tic = ax.text(0, rows - row, f"  {n}", transform=t, ha="left")

    plt.savefig(out, bbox_inches="tight")

    plt.close()
